class Model:
    """
    Model interface for each used data science model.
    Each model need to implement a @method: fit and a predict method
    """

    def fit(self, *args, **kwargs):
        raise NotImplementedError('The fit method needs to be implemented by each model')

    def predict(self, *args, **kwargs):
        raise NotImplementedError('The predict method needs to be implemented by each model')


class PrepInput:
    """
    PrepInput interface for each used data science model.
    Each PrepInput needs to implement fit, transform and fit_transform methods
    """

    def fit(self, *args, **kwargs):
        raise NotImplementedError('The fit method needs to be implemented by each model')

    def transform(self, *args, **kwargs):
        raise NotImplementedError('The transform method needs to be implemented by each model')

    def fit_transform(self, *args, **kwargs):
        raise NotImplementedError('The fit_transform method needs to be implemented by each model')



import yaml


def load_from_yaml(yaml_file, key=None):
    with open(yaml_file, 'rb') as fr:
        yaml_content = yaml.safe_load(fr)
    if key is None:
        return yaml_content
    return yaml_content[key]



from sagemaker import Session
from sagemaker.transformer import Transformer
from sbdscommons.constants import BUCKET
from sbdscommons.utils.yaml_utils import load_from_yaml

from .base import Model
from .base import PrepInput


class SbModel:
    separator = '-'

    def __init__(self,
                 model_name: str,
                 model: Model,
                 input_prep: PrepInput,
                 model_version: str,
                 bucket: str = BUCKET
                 ):
        self.model = model
        self.model_id = f'{model_name}{self.separator}{model_version}'
        self.input_prep = input_prep
        self.model_name = model_name
        self.input_train_path = f'{model_name}/input_data/train'
        self.input_predict_path = f'{model_name}/input_data/predict'
        self.output_path = f'{model_name}/output_data'
        self.model_path = f'{model_name}/model'
        self.bucket = bucket

    def __repr__(self):
        return self.model_id

    @staticmethod
    def load_model_params(params_path):
        ops_dict = load_from_yaml(params_path, key='options_dict')
        qry_dict = load_from_yaml(params_path, key='query_dict')
        return qry_dict, ops_dict

    def train(self, train_input):
        transformed_input = self.input_prep.transform(train_input)
        self.model.fit(transformed_input)

    def predict(self, predict_input):
        transformed_input = self.input_prep.transform(predict_input)
        return self.model.predict(transformed_input)

    def unload_for_predit(self):
        pass

    def unload_for_predit(self):
        pass

    def batch_transform_job(self):
        # CREATE A batch prediction job
        data_location = f's3://{self.bucket}/{self.output_path}'

        clf_trans = Transformer(
            base_transform_job_name='Batch-Transform',
            model_name=self.model_id,  # insert model name here
            instance_count=2,
            instance_type='ml.m5.2xlarge',
            output_path=data_location,
            sagemaker_session=Session(),
            max_payload=100
        )

        # Get some input data for the batch prediction job
        predict_data_key = f'{self.input_predict_path}/model_input_prediction000.csv'
        predict_data_location = f's3://{self.bucket}/{predict_data_key}'
        # start the prediction job
        clf_trans.transform(
            predict_data_location,
            content_type='text/csv',
            split_type='Line'
        )

        # wait until prediction job is completed
        clf_trans.wait()



import pickle
from io import StringIO

import flask
import pandas as pd
from flask import Response
from sbcommons.logging.lambda_logger import get_logger

from .model.model import SbModel

logger = get_logger(__name__)


def load_model() -> SbModel:
    with open('/opt/ml/model/sb_model.pickle', 'rb') as handle:
        sb_model = pickle.load(handle)
        logger.info(f'Loaded model {sb_model} using pickle on docker')
        return sb_model


def ping() -> Response:
    """
    Determine if the container is healthy by running a sample through the algorithm.
    """
    sb_model = load_model()
    predict_input_sample = pd.read_csv('tests/data/sample_model_input_prediction.csv',
                                       sep='|',
                                       quotechar='"')

    try:
        sb_model.predict(predict_input_sample)
        return Response(response='{"status": "ok"}', status=200, mimetype='application/json')
    except Exception as e:
        logger.exception(f'Unknown exception of type {e.__class__.__name__} occurred.')
        return Response(response='{"status": "error"}', status=500, mimetype='application/json')


def execution_parameters() -> Response:
    """
    SageMaker sets environment variables
    SAGEMAKER_MAX_PAYLOAD_IN_MB is set to the largest size payload that will be sent to the container via HTTP.

    SAGEMAKER_BATCH_STRATEGY will be set to SINGLE_RECORD when the container will be sent a single record per call to invocations and MULTI_RECORD when the container will get as many records as will fit in the payload.

    SAGEMAKER_MAX_CONCURRENT_TRANSFORMS is set to the maximum number of /invocations requests that can be opened simultaneously.

    """
    logger.info('execution-parameters function')
    return Response(response='{"status": "ok"}', status=200, mimetype='application/json')


def predict() -> Response:
    """
    Do an inference on a single batch of data.
    """
    if flask.request.content_type == 'text/csv':
        logger.info('[PREDICT] Reading input data.')
        predict_data = flask.request.data.decode('utf-8')
        buffer = StringIO()
        buffer.write(predict_data)
        buffer.seek(0)
        predict_input = pd.read_csv(buffer, sep='|', quotechar='"')
    else:
        return flask.Response(response='This predictor only supports CSV data', status=415,
                              mimetype='text/plain')

    logger.info('[PREDICT] Loading model...')
    sb_model = load_model()
    logger.info(f'[PREDICT] Using model: {SbModel} on the dataset of size: {predict_input.shape}')
    predict_data = sb_model.predict(predict_input)
    logger.info(f'[PREDICT] Done. Uploading dataset of size {predict_data.shape}.')
    out = StringIO()
    predict_data.to_csv(out, sep='|', quotechar='"', index=False)
    results_str = out.getvalue()
    return Response(response=results_str, status=200, mimetype='text/csv')



# ==========================
# AWS CONSTANTS
# ==========================
import os

ENV = 'prod'
ROLE = 'arn:aws:iam::485206745971:role/service-role/AmazonSageMaker-ExecutionRole-20190528T144422'
ACCOUNT = '485206745971'
REGION = 'eu-north-1'
BUCKET = f'sb-data-sc-{ENV}'

# Absolute path to folder with model params
PARAMS_FOLDER = f'{os.path.dirname(os.path.abspath(__file__))}/params'

# Model paths
INPUT_TRAIN_PATH_FMT = '{MODEL_NAME}/input_data/train'
INPUT_PREDICT_PATH_FMT = '{MODEL_NAME}/input_data/predict'
OUTPUT_PATH_FMT = '{MODEL_NAME}/output_data'
MODEL_PATH_FMT = '{MODEL_NAME}/model'



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="sbdscommons",
    version="0.0.8",
    author="Olof Nilsson",
    author_email="olof.nilsson@snusbolaget.se",
    description="Packages shared between sb data science projects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Snusbolaget/data-science",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True
)



import os
import pickle

import pandas as pd

from .model.model import SbModel


def build(sb_model: SbModel):
    # Get all input files downloaded
    train_path = '/opt/ml/input/data/train'
    input_files = [os.path.join(train_path, file) for file in os.listdir(train_path)]

    if len(input_files) == 0:
        raise Exception(f'There are no files in {train_path}. This usually indicates that the '
                        f'channel ("train") was incorrectly specified, the data specification in '
                        f'S3 was incorrectly specified or the role specified does not have '
                        f'permission to access the data.')

    # Take the set of files and read them all into a single pandas dataframe
    raw_data = [pd.read_csv(path, compression='gzip', sep='|', quotechar='"') for path in
                input_files]
    train_input = pd.concat(raw_data)
    sb_model.train(train_input)

    with open('sb_model.pickle', 'wb') as handle:
        pickle.dump(sb_model, handle)

    # Copy output model to the sagemaker model path
    os.system("cp sb_model.pickle /opt/ml/model/sb_model.pickle")
